from . import compiler


__version__ = '0.1.4'


def module(sqlpath):
    """
    Compiles a set of SQL files in the directory specified by sqlpath, and
    returns a module. The module contains a function for each named query
    found in the files.

        # create a module from sql files on disk
        queries = pugsql.module('path/to/sql/files')

        # connect to the database and use the sql queries as functions
        queries.connect(connection_string)
        queries.update_username(user_id=42, username='mcfunley')

    The results of this function are cached, so multiple calls giving the same
    sqlpath are safe and return the same module object.
    """
    return compiler.module(sqlpath)


__all__ = ['__version__', 'module',]



from . import lexer, statement
from itertools import takewhile


def parse(pugsql):
    stream = lexer.lex(pugsql)
    leading_comments = list(takewhile(lambda x: x[0] == 'C', stream))
    rest = stream[len(leading_comments):]

    cpr = parse_comments(leading_comments)
    sql = '\n'.join(cpr['unconsumed'] + [q for _, q in rest])

    return statement.Statement(
        name=cpr['name'],
        sql=sql,
        doc=cpr['doc'],
        result=cpr['result'])


def parse_comments(comments):
    cpr = {
        'name': None,
        'result': statement.Raw(),
        'doc': None,
        'unconsumed': [],
    }

    for _, c in comments:
        toks = lexer.lex_comment(c)
        if not toks:
            cpr['unconsumed'].append(c)
        elif toks['keyword'] == 'name':
            consume_name(cpr, toks)
        elif toks['keyword'] == 'result':
            consume_result(cpr, toks)
        else:
            cpr['unconsumed'].append(c)

    return cpr


def consume_result(cpr, tokens):
    if len(tokens['rest']) == 0:
        raise Exception('TODO')
    set_result(cpr, tokens['rest'])


def consume_name(cpr, tokens):
    # TODO deal with no name
    name, keywords = lexer.lex_name(tokens['rest'])
    cpr['name'] = name

    if len(keywords) == 0:
        return

    set_result(cpr, keywords[0])


def set_result(cpr, keyword):
    if keyword == ':one' or keyword == ':1':
        cpr['result'] = statement.One()
    elif keyword == ':many' or keyword == ':*':
        cpr['result'] = statement.Many()
    elif keyword == ':affected' or keyword == ':n':
        cpr['result'] = statement.Affected()
    elif keyword != ':raw':
        raise Exception('todo')



import re


def lex(pugsql):
    return [categorize(l) for l in pugsql.splitlines()]


def categorize(line):
    line = line.strip()
    return ('C', line) if line.startswith('--') else ('Q', line)


def lex_comment(c):
    m = re.match(
        r'--(?P<leading_whitespace>\s+)'
        r'\:(?P<keyword>[^ ]+)'
        r'\s+(?P<rest>.*)', c)
    return m.groupdict() if m else None


def lex_name(nameline):
    name, *rest = re.split(r'\s+', nameline.strip())
    return name, [k for k in rest if k.startswith(':')]



from . import parser
from glob import glob
import os
from sqlalchemy import create_engine


class Module(object):
    """
    Holds a set of SQL functions loaded from files.
    """

    def __init__(self, sqlpath):
        """
        Loads functions found in the *sql files specified by sqlpath into
        properties on this object.

        The named sql functions in files should be unique.
        """
        if not os.path.isdir(sqlpath):
            raise ValueError('Directory not found: %s' % sqlpath)

        self.sqlpath = sqlpath
        self._statements = {}

        for sqlfile in glob(os.path.join(self.sqlpath, '*sql')):
            with open(sqlfile, 'r') as f:
                pugsql = f.read()
            s = parser.parse(pugsql)

            if hasattr(self, s.name):
                raise ValueError(
                    'Error loading %s - a SQL function named %s was already '
                    'defined.' % (
                        sqlfile, s.name))

            setattr(self, s.name, s)
            self._statements[s.name] = s

    def connect(self, connstr):
        """
        Sets the connection string for SQL functions on this module.

        See https://docs.sqlalchemy.org/en/13/core/engines.html for examples of
        legal connection strings for different databases.
        """
        self.set_engine(create_engine(connstr))

    def set_engine(self, engine):
        """
        Sets the SQLAlchemy engine for SQL functions on this module. This can
        be used instead of the connect method, when more customization of the
        connection engine is desired.

        See also: https://docs.sqlalchemy.org/en/13/core/connections.html
        """
        for s in self._statements.values():
            s.set_engine(engine)


modules = {}


def module(sqlpath):
    """
    Compiles a new Module or returns a cached one. Use the pugsql.module
    instead of this one.
    """
    global modules
    if sqlpath not in modules:
        modules[sqlpath] = Module(sqlpath)
    return modules[sqlpath]



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['pugsql']

package_data = \
{'': ['*']}

install_requires = \
['sqlalchemy>=1.3,<2.0']

setup_kwargs = {
    'name': 'pugsql',
    'version': '0.1.4',
    'description': 'PugSQL is an anti-ORM that facilitates interacting with databases using SQL in files.',
    'long_description': None,
    'author': 'Dan McKinley',
    'author_email': 'mcfunley@gmail.com',
    'url': 'https://pugsql.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.4',
}


setup(**setup_kwargs)



from sqlalchemy.sql import text


class Result(object):
    def transform(self, r):
        raise NotImplementedError()


class One(Result):
    def transform(self, r):
        return { k: v for k, v in zip(r.keys(), r.first()) }


class Many(Result):
    def transform(self, r):
        ks = r.keys()
        return ({ k: v for k, v in zip(ks, row)} for row in r.fetchall())


class Affected(Result):
    def transform(self, r):
        return r.rowcount


class Raw(Result):
    def transform(self, r):
        return r


class Statement(object):
    def __init__(self, name, sql, doc, result):
        if not name:
            raise ValueError('Statement must have a name.')

        if sql is None:
            raise ValueError('Statement must have a SQL string.')
        sql = sql.strip()
        if not len(sql):
            raise ValueError('SQL string cannot be empty.')

        if not result:
            raise ValueError('Statement must have a result type.')

        self.name = name
        self.sql = sql
        self.doc = doc
        self.result = result
        self.engine = None

    def set_engine(self, engine):
        self.engine = engine

    def __call__(self, **params):
        if self.engine is None:
            raise RuntimeError(
                'No connection engine is configured. Pass a connection string '
                "to the module's connect method, or pass a SQLAlchemy engine "
                'to the set_engine method.')

        r = self.engine.execute(text(self.sql), **params)
        return self.result.transform(r)
